import gc
import weakref

import numpy as np

from . import logger

from lunavl.sdk.image_utils.image import MemoryResidence, VLImage  # isort: split


class PoolFrame(VLImage):
    """VLImage wrapper to support weak reference"""

    __slots__ = ("__weakref__",)


def catchFrame(imageRef: weakref.ReferenceType, framePool: list[PoolFrame]):
    """Catch core image before garbage collection"""
    frame = PoolFrame(imageRef())
    framePool.append(frame)


def iteratePool(framePool: list[PoolFrame], poolSize: int, width: int, height: int, memoryResidence: MemoryResidence):
    """Pool iteration"""
    while True:
        while framePool:
            frame = framePool.pop(0)
            finalizer = weakref.finalize(frame, catchFrame, weakref.ref(frame.coreImage), framePool)
            finalizer.atexit = False
            yield frame
        if poolSize > 0:
            logger.warning("Decoder frame pool cache miss")
        yield VLImage(np.ndarray(dtype=np.uint8, shape=(height, width, 3)), memoryResidence=memoryResidence)


class FramePool:
    """
    Decoder frame pool

    Frame pool, if enabled, manages access to preallocated frames in memory (ram or vram).
    Allocated images are provided to the client and returned to the frame pool when they are no longer in use.
    If the client starts using more frames than were allocated, the frame pool creates new frames and never
    returns them back to the pool
    """

    def __init__(self, size: int, memoryResidence: MemoryResidence, width: int, height: int):
        self._poolSize = size
        self.memoryResidence = memoryResidence
        self.width = width
        self.height = height

        framePool = [
            PoolFrame(np.ndarray(dtype=np.uint8, shape=(height, width, 3)), memoryResidence=memoryResidence)
            for _ in range(self._poolSize)
        ]

        self.pool = iteratePool(
            framePool=framePool,
            poolSize=self._poolSize,
            width=self.width,
            height=self.height,
            memoryResidence=memoryResidence,
        )

    def reset(self, width: int, height: int, memoryResidence: MemoryResidence):
        """Reset frame pool with new resolution"""
        self.width = width
        self.height = height
        self.memoryResidence = memoryResidence

        framePool = [
            PoolFrame(np.ndarray(dtype=np.uint8, shape=(height, width, 3)), memoryResidence=memoryResidence)
            for _ in range(self._poolSize)
        ]
        self.pool = iteratePool(
            framePool=framePool,
            poolSize=self._poolSize,
            width=self.width,
            height=self.height,
            memoryResidence=memoryResidence,
        )

    def close(self):
        """Execute gc"""
        self.pool = None
        gc.collect()
