"""
Module contains monitoring point protocol
"""

from abc import abstractmethod
from contextlib import contextmanager
from datetime import datetime
from time import perf_counter
from typing import Generator, Protocol


class MonitoringPoint(Protocol):
    """
    Protocol for a monitoring point.
    """

    # event time
    eventTime: float
    # series
    name: str

    @abstractmethod
    def getFields(self) -> dict:
        """Get monitoring fields"""
        raise NotImplementedError

    @abstractmethod
    def getTags(self) -> dict:
        """Get monitoring tags"""
        raise NotImplementedError


class Storage(Protocol):
    """
    Protocol for a point storage
    """

    @contextmanager
    @abstractmethod
    def popValues(self):
        """Get and pop values from storage."""
        raise NotImplementedError

    def append(self, point: MonitoringPoint):
        """Add point to storage"""
        raise NotImplementedError


class FramesPoint:
    """
    Frames monitoring point

    Attributes:
        eventTime: event time timestamp
        collectionTime: loading time (default=1 second)
        skippedCnt: number of skipped frames
        decodedCnt: number of decoded frames
        _startTime: start time of counter event task
        _stopTime: stop time of counter event task
    """

    name: str = "frames"

    def __init__(self, eventTime: float | None = None, collectionTime: float = 1.0):
        self.collectionTime = collectionTime
        self.eventTime = eventTime if eventTime is not None else datetime.now().timestamp()
        self.skippedCnt = 0
        self.decodedCnt = 0
        self._startTime = int(perf_counter())
        self._stopTime = self._startTime + 1

    def __str__(self):
        return str(self.getFields())

    def getFields(self) -> dict:
        """Get monitoring fields"""
        return {
            "collection_time": self.collectionTime,
            "num_skipped_frames": self.skippedCnt,
            "num_decoded_frames": self.decodedCnt,
        }

    def getTags(self) -> dict:
        """Get monitoring tags"""
        return {}


class FramesStorage:
    """
    Container for monitoring points

    Attributes:
        name: storage name
        _values: map - keys - series name, values - list monitoring points
        _lastPoint: info about last unfilled point
    """

    def __init__(self, name: str):
        self._values: dict[str, list[MonitoringPoint]] = {}
        self.name = name
        self._lastPoint: FramesPoint | None = None

    @contextmanager
    def popValues(self) -> Generator[dict[str, list[MonitoringPoint]], None, None]:
        """pop points from the storage"""
        # check the last skip event for readiness. Add it to the values if success
        if self._lastPoint is not None and self._lastPoint._stopTime <= perf_counter():
            self.append(self._lastPoint)
            self._lastPoint = None
        yield self._values
        # if correct processing points, remove them from the store
        self._values = {}

    def append(self, point: MonitoringPoint) -> None:
        """Add point to the storage"""
        self._values.setdefault(point.name, []).append(point)

    @property
    def _currentPoint(self) -> FramesPoint:
        """Get or create current monitoring point"""
        if self._lastPoint is None or self._lastPoint._stopTime <= perf_counter():
            newPoint = FramesPoint()
            if self._lastPoint is not None:
                self.append(self._lastPoint)
            self._lastPoint = newPoint
        return self._lastPoint

    def skipFrame(self):
        """Trigger a skip frame event"""
        self._currentPoint.skippedCnt += 1

    def countFrame(self):
        """Trigger a skip frame event"""
        self._currentPoint.decodedCnt += 1


# The storage contains an frames info
FRAMES_MONITORING = FramesStorage("frames")
