class TargetFrame:
    """The class allows you to determine whether a frame is a target based on its timestamp

    You pass the target fps to the initializer, and then call a method that returns a boolean value about whether
    the frame is targeted or not.

    Attributes:
        _targetFps: target fps
        _frameInterval: interval in seconds between two target frames
        _nextFrameTs: timestamp of next target frame
        _lastFrameTs: timestamp of last target frame
    """

    __slots__ = ("_targetFps", "_frameInterval", "_nextFrameTs", "_lastFrameTs")

    def __init__(self, targetFps: float):
        if targetFps <= 0:
            raise ValueError("targetFps must be greater than 0")

        self._targetFps: float
        self._frameInterval: float
        self._nextFrameTs: float = 0.0
        self._lastFrameTs: float = 0.0

        self.targetFps = targetFps

    def __call__(self, frameTs: float) -> bool:
        """Determines whether a frame is targeted

        If the first frame was transmitted, then its Ts will be the initial one from which other Ts will be calculated.
        If the obtained ts greatly exceeds the target, then a recalculation will be performed.

        Args:
            frameTs (float): frame timestamp

        Returns:
            bool: target frame or not
        """
        accumulationError = 1e-6  # frame accumulation error
        if self._nextFrameTs == 0.0:  # init frame
            self._nextFrameTs = frameTs + self._frameInterval
            return True
        if frameTs + accumulationError >= self._nextFrameTs:
            # Handle large gaps in frameTs by skipping missed frames
            missedFrames = int((frameTs + accumulationError - self._nextFrameTs) // self._frameInterval) + 1
            self._lastFrameTs = self._nextFrameTs
            self._nextFrameTs += missedFrames * self._frameInterval
            return True
        return False

    @property
    def targetFps(self) -> float:
        return self._targetFps

    @targetFps.setter
    def targetFps(self, fps: float) -> None:
        """Set new target fps

        If _nextFrameTs has already been initialized, then the parameters for the next frame will be recalculated.

        Args:
            fps (float): new target fps

        """
        if fps <= 0:
            raise ValueError("targetFps must be greater than 0")

        self._targetFps = fps
        self._frameInterval = 1 / self._targetFps
        if self._lastFrameTs:
            self._nextFrameTs = self._lastFrameTs + self._frameInterval
        elif self._nextFrameTs:
            self._lastFrameTs = self._nextFrameTs
            self._nextFrameTs += self._frameInterval

    @property
    def nextFrameTs(self) -> float:
        return self._nextFrameTs

    @property
    def lastFrameTs(self) -> float:
        return self._lastFrameTs
