import asyncio
import dataclasses
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager, contextmanager
from pathlib import Path
from typing import AsyncGenerator, Callable, Generator, Literal, SupportsIndex, Tuple, cast

import numpy as np
import PyFsdkInterop as psi
import PyNvCodec as nvc
from cuda import cuda, cudart
from lunavl.sdk.image_utils.image import MemoryResidence, VLImage

from . import logger
from .decode_queue import DecodeQueue
from .enums import DecoderState, DeviceClass, ErrorAction, OrientationStrategy, PTSPolicy, VideoSource
from .exceptions import (
    DeviceIsNotAvailableException,
    FileNotFoundException,
    InvalidDataException,
    PTSNotFound,
    SeekUnsupportedError,
    UnsupportedPixelFormatException,
    UnsupportedUserAngle,
    UnsupportedVideoFormatException,
    UnsupportedVideoInfoAngle,
    VideoDecodingException,
    VideoDecodingTimeout,
)
from .frame_pool import FramePool, PoolFrame
from .structs import ORIENTATION_MAP, VideoData, VideoSegment
from .target_frame import TargetFrame

GPU_ID = 0


@dataclasses.dataclass
class ErrorHandling:
    """Error handling options"""

    action: ErrorAction = ErrorAction.fail  # action when an error occurs
    collectProcessing: bool = False  # collect or not success processing parts. Not use for infinity stream


class VideoDecoder:
    """
    Video decoder

    preferredDecoderDevice (DeviceClass): preferred device for video decoding
    executor (ThreadPoolExecutor): executor for performing async decoding

    Attributes:
        videoPath (str): path to video
        seek (float): video seek timestamp
        _packetData (PacketData): ffmpeg packet data
        _loop:  python event loop
        _decoder: video decoder
        _lastDecodedFrame (VideoData): last decoded frame
        _state: internal decoder state
        _avgFramerate: average frames per second
        decodedFrameCount: decoded frame count
        _videoTimeBase: video TimeBase from ffmpeg
        _startTime: video start time from ffmpeg
        _decodingMethod: frame decoding method
        _conversionMethod: frame conversion method
        _framePool: decoder frame pool
        _cvtCtx: color conversion context
        _nvCvt: frame converter
        _originStartTime: video start time from ffmpeg
        _externalStartTime: user defined video start time.
           If None, _externalStartTime will be used from video or 0.  If  not None frame pts will be counted from
           this time.
        _startTime: final video start time, calculated from _originStartTime, _externalStartTime, and video data
        _timeShift: frame pts time shift
        _errorHandling: error handler options
                fail = 1  - raise VideoDecodingException
                stop = 2  - graceful stop decoding and close stream
                reconnect - a reconnect to the stream in error case , if reconnect is failed then
                            VideoDecodingEsception will be raised
        _processingParts: success processing parts
        _errorCount: number of errors when decoding video
        _lastTS: last calculated package timestamp
        _ptsPolicy: pts source
        _source: a data about video source
        _useMonitoring: use or no monitoring
        _decodeQueue: queue for storing raw decoded frames
        _queuePutMethod: method of adding to queue: wait until there is space, or add immediately with offset
        _decodeToQueueTask: task performing decoding in queue
        _decodedTrueFrameCount: the number of all decoded frames, including those skipped
    """

    # preferred device for decoding
    preferredDecoderDevice: DeviceClass
    # thread pool executor for decoding and converting frames
    executor: ThreadPoolExecutor | None = None
    # frame pool size
    poolSize = 0
    # ffmpeg tread count for decoding
    _ffmpegTreadCount: int | None = None
    # stream decode queue size
    streamDecodeQueueSize: int = 50
    # orientation strategy: static or using video info
    _orientationStrategy: OrientationStrategy
    # rotation angle
    _angle: Literal[0, 90, 180, 270] | None = None

    def __init__(
        self,
        videoPath: str | Path,
        startTime: float | None = None,
        errorHandling: ErrorHandling | None = None,
        ptsPolicy: PTSPolicy = PTSPolicy.auto,
        source: VideoSource | None = None,
        seek: float | None = None,
        useMonitoring: bool = False,
        orientationStrategy: OrientationStrategy = OrientationStrategy.static,
        angle: Literal[0, 90, 180, 270] = 0,
        targetFps: float | None = None,
    ):
        self._orientationStrategy = orientationStrategy
        self._angle = angle if self._orientationStrategy == OrientationStrategy.static else None
        self.videoPath = videoPath
        self.seek: float = seek if seek else 0
        self._loop = asyncio.get_running_loop()
        self._packetData = nvc.PacketData()
        self._decoder: nvc.PyDecoder = None
        self._decodingMethod: Callable[
            [np.ndarray[SupportsIndex, np.dtype[np.uint8]] | nvc.Surface],
            Tuple[np.ndarray[SupportsIndex, np.dtype[np.uint8]] | nvc.Surface, nvc.TaskExecInfo],
        ]
        self._conversionMethod: Callable[[np.ndarray[SupportsIndex, np.dtype[np.uint8]], VLImage], Tuple[bool, VLImage]]
        self.decodedFrameCount = 0
        self._lastDecodedFrame: VideoData | None = None
        self._decoderKwargs: dict[str, str] = {}
        self._state = DecoderState.initiated
        self._iterator = None
        self._avgFramerate: float = 0.0
        self._timestampGetter: Callable[[], float] = self._ptsTimestamp
        self._originStartTime: float | None = None
        self._externalStartTime: float | None = startTime
        self._videoTimeBase: float | None = None
        self._innerFramePool: list[np.ndarray[SupportsIndex, np.dtype[np.uint8]] | nvc.Surface] = []
        self._innerFramePoolSize: int = 0
        self._cvtCtx: nvc.ColorspaceConversionContext
        self._nvCvt: nvc.PyFrameConverter | nvc.PySurfaceConverter
        self._framePool: FramePool
        self._innerPoolIterator: Generator
        self._errorHandling = ErrorHandling() if errorHandling is None else errorHandling
        self._timeShift: float = 0.0
        self._startTime: float = 0.0
        self._seeked = False if self.seek else True
        if self._errorHandling.collectProcessing:
            self._processingParts: None | list[VideoSegment] = []
        else:
            self._processingParts = None
        self._errorCount = 0
        self._lastTS = 0.0
        self._source = source
        self._ptsPolicy = ptsPolicy
        if self._source == VideoSource.stream and self._ptsPolicy == PTSPolicy.frameRate:
            raise ValueError("Source `stream` and pts policy `framerate` is incompatibility")
        if self._source == VideoSource.stream and self.seek:
            raise SeekUnsupportedError("Source `stream` and `seek` is incompatibility")
        if self._source == VideoSource.file and self._errorHandling.action == ErrorAction.reconnect:
            raise ValueError("Source `file` is not supported reconnect")
        self._decoderStopEvent = self._loop.create_future()
        self._device: Literal["gpu", "cpu"] | None = None
        self._cudaStream = None
        self._useMonitoring = useMonitoring
        # queue for storing encoded raw frames
        self._decodeQueue = DecodeQueue(
            maxsize=self.streamDecodeQueueSize if self._source == VideoSource.stream else 1,
            useMonitoring=self._useMonitoring,
            useLogger=True,
        )
        # putWithShift for streams or put for video
        self._queuePutMethod = (
            self._decodeQueue.putWithShift if self._source == VideoSource.stream else self._decodeQueue.put
        )
        self._decodeToQueueTask: asyncio.Task
        self._decodedTrueFrameCount = 0
        self._targetFps = targetFps
        self._targetFrameTraking: TargetFrame

        if self._targetFps is not None:
            self._targetFrameTraking = TargetFrame(self._targetFps)

    @classmethod
    def _initCuda(cls) -> bool:
        """
        Init cuda structures
        """
        # https://nvidia.github.io/cuda-python/module/cuda.html?highlight=cuinit#cuda.cuda.cuInit
        # Currently, the Flags parameter must be 0
        err, res = cudart.cudaRuntimeGetVersion()
        if err == cuda.CUresult.CUDA_SUCCESS:
            logger.info(f"runtime cuda version {res}")
        else:
            logger.info(f"cuda driver not found, detail: {err}")
            return False
        return True

    @classmethod
    async def initialize(
        cls,
        deviceClass: Literal["auto", "gpu", "cpu"],
        workerCount: int,
        poolSize: int = 0,
        ffmpegTreadCount: int | None = None,
    ) -> None:
        """Initialize video decoder"""
        cls.preferredDecoderDevice = DeviceClass[deviceClass]
        cls.executor = ThreadPoolExecutor(max_workers=workerCount)
        cls.poolSize = poolSize
        cls._ffmpegTreadCount = ffmpegTreadCount
        if deviceClass in ("auto", "gpu"):
            res = cls._initCuda()
            if not res:
                if deviceClass == "gpu":
                    raise RuntimeError("Failed to initialize CUDA device")
                logger.info("Video decoding will be to compute on cpu")
                cls.preferredDecoderDevice = DeviceClass["cpu"]

    def _createGPUDecoder(self, timeout: int = 5000) -> nvc.PyDecoder:
        """Create gpu decoder"""
        nvDecoder = nvc.PyDecoder(str(self.videoPath), opts={"timeout": str(timeout)}, gpu_id=GPU_ID)
        return nvDecoder

    def _createCPUDecoder(self, timeout: int = 5000) -> nvc.PyDecoder:
        """Create cpu decoder"""
        ffmpegArgs = {"timeout": str(timeout)}
        if self._ffmpegTreadCount:
            ffmpegArgs["threads"] = str(self._ffmpegTreadCount)
        nvDecoder = nvc.PyDecoder(str(self.videoPath), opts=ffmpegArgs, gpu_id=-1)
        return nvDecoder

    def _createDecoder(self, decoderDevice: DeviceClass, timeout: int = 5000) -> nvc.PyDecoder:
        """
        Create video decoder

        Args:
            decoderDevice: device for performing decoding.
            timeout: decoder creation timeout

        Returns:
            Video decoder instance
        Raises:
            InvalidDataException: if data related error occurs
        """
        try:
            if decoderDevice == DeviceClass.gpu:
                nvDecoder = self._createGPUDecoder(timeout=timeout)
                self._device = "gpu"
            elif decoderDevice == DeviceClass.cpu:
                nvDecoder = self._createCPUDecoder(timeout=timeout)
                self._device = "cpu"
            else:
                raise RuntimeError(f"Unknown device class {decoderDevice}")
        except RuntimeError as err:
            message = str(err)
            if "Invalid data" in message:
                raise InvalidDataException() from err
            if "unknown filter" in message or "Can't find codec" in message:
                raise UnsupportedVideoFormatException("Unsupported format")
            if "cuInit not found" in message:
                raise DeviceIsNotAvailableException("GPU not found") from err
            if "Can't open source file" in message:
                raise FileNotFoundException(f"Video file not found: {self.videoPath}") from err
            raise
        except ValueError as err:
            raise InvalidDataException() from err
        return nvDecoder

    def _makeVRMBuffer(self, width: int, height: int) -> nvc.Surface:
        """Create buffer for decoding into VRAM memory"""
        return nvc.Surface.Make(self._decoder.Format(), width, height, GPU_ID)

    def _makeRAMBuffer(self, width: int, height: int) -> np.ndarray:
        """Create buffer for decoding into ram memory"""
        return np.ndarray(dtype=np.uint8, shape=(0))

    def _initFramePool(self, width: int, height: int, memoryResidence: MemoryResidence):
        """Initialize frame pool"""
        self._framePool = FramePool(self.poolSize, memoryResidence, width, height)
        self._initConverter(width, height, memoryResidence)

        if self.poolSize > 0:

            if memoryResidence == MemoryResidence.CPU:
                makeBuffer = self._makeRAMBuffer
            else:
                makeBuffer = self._makeVRMBuffer
            self._innerFramePool = [makeBuffer(width, height) for _ in range(self.poolSize)]
            self._innerFramePoolSize = self.poolSize

            def iterateInnerPool():
                while True:
                    while self._innerFramePool:
                        yield self._innerFramePool.pop(0)
                    logger.warning("Decoder inner frame pool cache miss")
                    yield makeBuffer(width, height)

        else:

            def iterateInnerPool():
                while True:
                    if memoryResidence == MemoryResidence.CPU:
                        yield self._makeRAMBuffer(width, height)
                    else:
                        yield self._makeVRMBuffer(width, height)

        self._innerPoolIterator = iterateInnerPool()

    @contextmanager
    def _innerPool(self):
        """Catch buffer back to pool if needed"""
        buffer = next(self._innerPoolIterator)
        try:
            yield buffer
        finally:
            if len(self._innerFramePool) < self._innerFramePoolSize:
                self._innerFramePool.append(buffer)

    def _resetFramePool(self, width: int, height: int, memoryResidence: MemoryResidence):
        """Reset frame pool with new resolution"""
        self._framePool.reset(width, height, memoryResidence)
        self._initFramePool(width, height, memoryResidence)
        self._initConverter(width, height, memoryResidence)

    def _initConverter(self, width: int, height: int, memoryResidence: MemoryResidence):
        """Init decoder frame converter"""
        colorSpace = nvc.ColorSpace.BT_709
        colorRange = nvc.ColorRange.JPEG
        if (frameFormat := self._decoder.Format()) == nvc.PixelFormat.YUV420:
            colorSpace = nvc.ColorSpace.BT_601
        self.frameFormat = frameFormat
        self._cvtCtx = nvc.ColorspaceConversionContext(colorSpace, colorRange)
        try:
            if memoryResidence == MemoryResidence.GPU:
                self._nvCvt = nvc.PySurfaceConverter(width, height, self.frameFormat, nvc.PixelFormat.RGB, GPU_ID)
            else:
                self._nvCvt = nvc.PyFrameConverter(width, height, self.frameFormat, nvc.PixelFormat.RGB)
        except ValueError as err:
            if err.args[0].startswith("ConvertFrameFfmpeg: Unsupported pixel format conversion"):
                raise UnsupportedPixelFormatException(err.args[0])
            raise

    def _decodeSurface(self, buffer: nvc.Surface) -> tuple[nvc.Surface, nvc.TaskExecInfo]:
        """Decode frame to gpu memory"""
        if self._seeked:
            res = self._decoder.DecodeSingleSurface(buffer, self._packetData, **self._decoderKwargs)
        else:
            seekCtx = nvc.SeekContext(seek_ts=self.seek)
            res = self._decoder.DecodeSingleSurface(
                surf=buffer, pkt_data=self._packetData, seek_ctx=seekCtx, **self._decoderKwargs
            )
            self._seeked = True
        return buffer, res[1]

    def _convertSurface(self, frame: nvc.Surface, outFrame: VLImage) -> Tuple[bool, VLImage]:
        """Convert frame from gpu"""
        with psi.AsSurface(outFrame.coreImage) as surfaceView:
            success = self._nvCvt.Execute(frame, surfaceView, self._cvtCtx)
        return success, outFrame

    def _decodeFrame(
        self, buffer: np.ndarray[SupportsIndex, np.dtype[np.uint8]]
    ) -> tuple[np.ndarray[SupportsIndex, np.dtype[np.uint8]], nvc.TaskExecInfo]:
        """Decode frame to ram"""
        if self._seeked:
            isNewFrame, status = self._decoder.DecodeSingleFrame(buffer, self._packetData, **self._decoderKwargs)
        else:
            seekCtx = nvc.SeekContext(seek_ts=self.seek)
            isNewFrame, status = self._decoder.DecodeSingleFrame(
                buffer, self._packetData, seek_ctx=seekCtx, **self._decoderKwargs
            )

            self._seeked = True
        return buffer, status

    def _convertFrame(
        self, frame: np.ndarray[SupportsIndex, np.dtype[np.uint8]], outFrame: VLImage
    ) -> Tuple[bool, VLImage]:
        """Convert frame from ram"""
        success = self._nvCvt.Execute(frame, outFrame.asNPArray(), self._cvtCtx)
        return success, outFrame

    async def _decodeSingleFrame(self) -> tuple[VLImage | None, nvc.TaskExecInfo]:
        """Decode a frame"""
        with self._innerPool() as buffer:
            frame, status = await self._loop.run_in_executor(self.executor, self._decodingMethod, buffer)
            if status == nvc.TaskExecInfo.RES_CHANGE:
                self._resetFramePool(self._decoder.Width(), self._decoder.Height(), self._framePool.memoryResidence)
                res = await self._decodeSingleFrame()
                return res
        return frame, status

    async def _decodeFrames(self) -> None:
        """Decodes frames and puts them in a queue.

        Performs:
        * Calculates the timestamp of each frame
        * Determines whether the frame is a target

        Puts the frame in the queue if the next frame has already been received or an error has occurred.
        This is necessary because the last frame needs to be synchronized with the end of decoding.

        """
        _frame, _status, _ts, _pool = None, None, None, None
        while True:
            frame, status = await self._decodeSingleFrame()

            if status == nvc.TaskExecInfo.SUCCESS:
                self._decodedTrueFrameCount += 1
                ts = self._timestampGetter()  # get timestamp for current frame
                self._lastTS = ts
                # it is always necessary to calculate the timestamp in order to have correct values ​​in the future
                if self._targetFps and not self._targetFrameTraking(ts):  # skip frame if needed
                    continue
                # get the current frame pool, this is necessary to synchronize decoding and iteration over frames
                pool = self._framePool.pool
                # put a frame in the queue because a new frame was received
                if _status is not None:
                    await self._queuePutMethod((_frame, _status, _ts, _pool))
                _frame, _status, _ts, _pool = frame, status, ts, pool
            else:
                if _status is not None:
                    await self._queuePutMethod((_frame, _status, _ts, _pool))
                await self._queuePutMethod((frame, status, None, None))
                break

    def getRotationAngle(self) -> int:
        """Get rotation angle"""
        if self._orientationStrategy == OrientationStrategy.static:
            angle = self._angle
            if angle not in ORIENTATION_MAP:
                raise UnsupportedUserAngle(f"Unsupported rotation angle from user-defined settings: {angle}")
        elif self._orientationStrategy == OrientationStrategy.videoInfo:
            displayRotation: float = self._decoder.DisplayRotation()
            try:
                angle = int(displayRotation)  # type: ignore
                if angle not in ORIENTATION_MAP:
                    raise ValueError
            except ValueError:
                raise UnsupportedVideoInfoAngle(f"Unsupported display rotation from video info: {displayRotation}")
        else:
            raise RuntimeError(f"Unsupported orientation strategy: {self._orientationStrategy}")
        return angle  # type: ignore

    def _rotateFrame(self, frame: VLImage) -> VLImage:
        """Rotate frame according due to orientation strategy"""
        angle = self.getRotationAngle()

        if angle == 0:
            return frame
        return VLImage.rotate(frame, ORIENTATION_MAP[angle])

    def _convertToRGB(
        self, frame: np.ndarray[SupportsIndex, np.dtype[np.uint8]] | nvc.Surface, outFrame: VLImage
    ) -> PoolFrame:
        """
        Convert frame to RGB format

        Args:
            frame: frame to convert

        Returns:
            Frame converted to RGB format
        """
        success, outFrame = self._conversionMethod(frame, outFrame)

        if not success:
            raise VideoDecodingException("Format converting error")
        return outFrame

    def _convertNRotate(
        self, frame: np.ndarray[SupportsIndex, np.dtype[np.uint8]] | nvc.Surface, outFrame: VLImage
    ) -> VLImage:
        """Convert to RGB and rotate frame"""
        convertedFrame = self._convertToRGB(frame=frame, outFrame=outFrame)
        rotatedFrame = self._rotateFrame(frame=convertedFrame)
        return rotatedFrame

    async def _createDecoderAsync(self, decoderDevice: DeviceClass, timeout: int = 5000) -> nvc.PyDecoder:
        """
        Create video decoder
        Args:
            decoderDevice: decoder device class
            timeout: decoder creation timeout
        Returns:
            decoder
        """
        logger.info(f"Create decoder, device: {decoderDevice.value}")

        return await self._loop.run_in_executor(self.executor, self._createDecoder, decoderDevice, timeout)

    async def createDecoder(self, device: DeviceClass = DeviceClass.auto, timeout: int = 5000) -> nvc.PyDecoder:
        """
        Create video decoder for device
        Args:
            device: device class
            timeout: decoder creation timeout
        Returns:
            decoder
        """

        if device == DeviceClass.auto:
            try:
                return await self._createDecoderAsync(decoderDevice=DeviceClass.gpu, timeout=timeout)
            except (RuntimeError, DeviceIsNotAvailableException, UnsupportedVideoFormatException) as err:
                logger.warning(f"Fail to initialize gpu decoder, fallback to cpu. Reason: {err}")
                return await self._createDecoderAsync(decoderDevice=DeviceClass.cpu, timeout=timeout)
        return await self._createDecoderAsync(decoderDevice=device, timeout=timeout)

    async def _initDecoder(self, device: DeviceClass | None = None, timeout: int = 5000):
        """
        Initialize decoding and conversion behavior
        Args:
            device: device to initialize, if None set, the self.preferredDecoderDevice will be used
            timeout: decoder timeout
        """
        if device is None:
            device = self.preferredDecoderDevice
        self._decoder = await self.createDecoder(device=device, timeout=timeout)
        width, height = self._decoder.Width(), self._decoder.Height()
        if self._device == "gpu":
            memoryResidence = MemoryResidence.GPU
            self._decodingMethod = self._decodeSurface
            self._conversionMethod = self._convertSurface
        else:
            memoryResidence = MemoryResidence.CPU
            self._decodingMethod = self._decodeFrame
            self._conversionMethod = self._convertFrame
        self._initFramePool(width, height, memoryResidence)

    async def _getDecoder(self, timeout: int = 5000) -> nvc.PyDecoder:
        """
        Get video decoder, create if not exists

        Args:
            timeout: decoder timeout
        Returns:
            Video decoder instance
        """
        if self._decoder is None:
            await self._initDecoder(timeout=timeout)
        return self._decoder

    async def _syncDecodeNextFrame(
        self,
    ) -> tuple[np.ndarray[SupportsIndex, np.dtype[np.uint8]] | None, nvc.TaskExecInfo]:
        """Get frame, perform format conversion to RGB format and return frame and status"""
        if self._decoder is None:
            raise RuntimeError("decoder does not initialize")
        try:
            frame, status = await self._decodeSingleFrame()
            outFrame = None
            if status == nvc.TaskExecInfo.SUCCESS:
                outFrame = next(self._framePool.pool)
                outFrame = await self._loop.run_in_executor(self.executor, self._convertNRotate, frame, outFrame)
        except RuntimeError as e:
            if "Decoded frame doesn't have PTS, can't seek." in str(e):
                raise SeekUnsupportedError(f"{self.videoPath}, video doesn`t support seek operation") from e
            else:
                raise e
        if status == nvc.TaskExecInfo.SUCCESS:
            self.decodedFrameCount += 1
            self._decodedTrueFrameCount += 1
        return outFrame, status

    async def _decodeNextFrame(
        self,
    ) -> tuple[np.ndarray[SupportsIndex, np.dtype[np.uint8]] | None, nvc.TaskExecInfo, float]:
        """Get frame from queue, perform format conversion to RGB format and return frame and status"""
        if self._decoder is None:
            raise RuntimeError("decoder does not initialize")
        try:
            frame, status, ts, pool = await self._decodeQueue.get()
            outFrame = None
            if status == nvc.TaskExecInfo.SUCCESS:
                outFrame = next(pool)
                outFrame = await self._loop.run_in_executor(self.executor, self._convertNRotate, frame, outFrame)
        except RuntimeError as e:
            if "Decoded frame doesn't have PTS, can't seek." in str(e):
                raise SeekUnsupportedError(f"{self.videoPath}, video doesn`t support seek operation") from e
            else:
                raise e
        if status == nvc.TaskExecInfo.SUCCESS:
            self.decodedFrameCount += 1
        return outFrame, status, ts

    def prepareSeek(self) -> None:
        """Prepare seek for use."""
        if self.seek < 0 and self.duration:
            self.seek = self.duration - abs(self.seek)
        elif self.seek < 0:
            raise SeekUnsupportedError(f"{self.videoPath} video doesn`t support negative seek timestamp.")

    async def prepare(self, timeout: int = 5000) -> bool:
        """
        Prepare internal structures and decode first frame. This function allows to split codec initialization +
        checking of ability with video and frame decoding.

        Args:
            timeout: decoder timeout
        Returns:
            false if video is empty
        """
        if self._state == DecoderState.prepared:
            return True
        elif self._state != DecoderState.initiated:
            raise RuntimeError(f"bad decoder state {self._state}")
        try:
            await self._getDecoder(timeout=timeout)
        except BaseException:
            self._errorCount += 1
            self._state = DecoderState.error
            raise
        if self.seek:
            self.prepareSeek()
        buffer, status = await self._syncDecodeNextFrame()
        if status == nvc.TaskExecInfo.END_OF_STREAM:
            self._state = DecoderState.completed
            logger.info("stream is closed")
            return False
        if status != nvc.TaskExecInfo.SUCCESS:
            if self.preferredDecoderDevice == DeviceClass.auto and self._device == "gpu":
                try:
                    await self._initDecoder(DeviceClass.cpu)
                except BaseException:
                    self._errorCount += 1
                    self._state = DecoderState.error
                    raise
                logger.warning(f"Decoding failed on gpu, fallback to cpu")
                self._packetData = nvc.PacketData()
            else:
                self._errorCount += 1
                self._state = DecoderState.error
                if status == nvc.TaskExecInfo.TIMEOUT:
                    raise VideoDecodingTimeout(status.name, frameNumber=self.decodedFrameCount, ts=1)
                raise VideoDecodingException(status.name, frameNumber=self.decodedFrameCount, ts=1)
            if self.seek:
                self.prepareSeek()
            buffer, status = await self._syncDecodeNextFrame()
            if status == nvc.TaskExecInfo.END_OF_STREAM:
                self._state = DecoderState.completed
                logger.info("stream is closed")
                return False
            if status != nvc.TaskExecInfo.SUCCESS:
                self._state = DecoderState.error
                self._errorCount += 1
                if status == nvc.TaskExecInfo.TIMEOUT:
                    raise VideoDecodingTimeout(status.name, frameNumber=self.decodedFrameCount, ts=1)
                raise VideoDecodingException(status.name, frameNumber=self.decodedFrameCount, ts=1)
        # setup persistence video data
        self._setUpFrameTimestampCalculator()
        ts = self._timestampGetter()
        self._lastDecodedFrame = VideoData(buffer, 1, ts, False)
        if self._targetFps is not None:
            self._targetFrameTraking(ts)  # set firts target frame
        self._state = DecoderState.prepared
        logger.info("success open stream")
        return True

    @asynccontextmanager
    async def decode(self):
        """Control async generator finalization"""
        decode = self._decode()
        try:
            yield decode
        finally:
            if decode.ag_running:
                await self.scheduleToStop()
            await decode.aclose()
            self._lastDecodedFrame = None
            self._framePool.close()

    async def scheduleToStop(self):
        """Schedule decode generator to stop iterating"""
        if self._state != DecoderState.decoding:
            return
        self._state = DecoderState.stoping
        await self._decoderStopEvent

    async def _reconnect(self):
        self._state = DecoderState.initiated
        self._decoder = None
        self._packetData = nvc.PacketData()
        if await self.prepare():
            logger.info("Success reconnect to stream")
            self._lastDecodedFrame = VideoData(
                self._lastDecodedFrame.buffer,
                self.decodedFrameCount,
                self._lastDecodedFrame.timestamp,
                self._lastDecodedFrame.lastFrame,
            )
            self._decodeToQueueTask = asyncio.create_task(self._decodeFrames())
            return True
        logger.info(f"Reconnect is failed")
        return False

    async def _decode(self) -> AsyncGenerator[VideoData, None]:
        """
        Decode video frame by frame
        """
        match self._state:
            case DecoderState.initiated:
                if not await self.prepare():
                    return
            case DecoderState.prepared:
                pass
            case _:
                raise RuntimeError(f"bad decoder state {self._state}")

        self._decodeToQueueTask = asyncio.create_task(self._decodeFrames())
        self._state = DecoderState.decoding
        try:
            while True:
                if self._state == DecoderState.stoping:
                    raise asyncio.CancelledError
                lastDecodedFrame: VideoData = cast(VideoData, self._lastDecodedFrame)
                buffer, status, ts = await self._decodeNextFrame()
                if status == nvc.TaskExecInfo.END_OF_STREAM:
                    logger.info("end of stream")
                    yield VideoData(lastDecodedFrame.buffer, lastDecodedFrame.number, lastDecodedFrame.timestamp, True)
                    break
                elif status != nvc.TaskExecInfo.SUCCESS:
                    logger.info(f"error decoding, {status.name}")
                    self._errorCount += 1
                    match self._errorHandling.action:
                        case ErrorAction.fail:
                            if status == nvc.TaskExecInfo.TIMEOUT:
                                raise VideoDecodingTimeout(
                                    status.name, frameNumber=self.decodedFrameCount, ts=lastDecodedFrame.timestamp
                                )
                            raise VideoDecodingException(
                                status.name, frameNumber=self.decodedFrameCount, ts=lastDecodedFrame.timestamp
                            )
                        case ErrorAction.stop:
                            if self._errorHandling.collectProcessing:
                                self._processingParts = [VideoSegment(self._startTime, lastDecodedFrame.timestamp)]
                            yield VideoData(
                                lastDecodedFrame.buffer, lastDecodedFrame.number, lastDecodedFrame.timestamp, True
                            )
                            break
                        case ErrorAction.reconnect:
                            if await self._reconnect():
                                continue
                            raise VideoDecodingException(
                                status.name, frameNumber=self.decodedFrameCount, ts=lastDecodedFrame.timestamp
                            )
                        case _:
                            if status == nvc.TaskExecInfo.TIMEOUT:
                                raise VideoDecodingTimeout(
                                    status.name,
                                    frameNumber=self.decodedFrameCount,
                                    ts=lastDecodedFrame.timestamp,
                                )
                            raise VideoDecodingException(
                                status.name, frameNumber=self.decodedFrameCount, ts=lastDecodedFrame.timestamp
                            )
                _lastDecodeData = VideoData(buffer, self.decodedFrameCount, ts, False)
                yield lastDecodedFrame
                self._lastDecodedFrame = _lastDecodeData
        except VideoDecodingException:
            self._state = DecoderState.error
            raise
        except Exception:
            self._errorCount += 1
            self._state = DecoderState.interrupted
            raise
        except asyncio.CancelledError:
            self._state = DecoderState.interrupted
            raise
        else:
            self._state = DecoderState.completed
            if self._errorHandling.collectProcessing:
                self._processingParts = [
                    VideoSegment(self._startTime, cast(VideoData, self._lastDecodedFrame).timestamp)
                ]
        finally:
            self._decoderStopEvent.set_result(True)
            self._decodeToQueueTask.cancel()
            logger.info(f"end decoding")

    @property
    def state(self) -> DecoderState:
        """Get current decoder state"""
        return self._state

    @property
    def lastDecodedFrame(self) -> VideoData | None:
        """Get last decoded frame"""
        return self._lastDecodedFrame

    def _oldframeRateTimestamp(self) -> float:
        """Get frame timestamp by framerate"""
        return (self.decodedFrameCount - 1) / self._avgFramerate + self._timeShift

    def _frameRateTimestamp(self) -> float:
        return (self._decodedTrueFrameCount - 1) / self._avgFramerate + self._timeShift

    def _setUpFramerateTimestampGetter(self):
        if self._avgFramerate == 0.0:
            self._avgFramerate = self._decoder.AvgFramerate()
        if self._externalStartTime is not None:
            self._timeShift = self._externalStartTime
        else:
            self._timeShift = 0
        self._startTime = 0
        self._timestampGetter = self._frameRateTimestamp
        self._ptsPolicy = PTSPolicy.frameRate

    def _setUpPTSTimestampGetter(self):
        """Set up package pts time as pts source"""
        self._videoTimeBase = self._decoder.Timebase()
        self._originStartTime = self._decoder.StartTime() or 0
        self._avgFramerate = self._decoder.AvgFramerate()
        if self._externalStartTime is None:
            self._timeShift = 0
            self._startTime = self._originStartTime
        else:
            if self._source == VideoSource.stream:
                self._timeShift = 0
            else:
                self._timeShift = self._externalStartTime - self._originStartTime
            self._startTime = self._externalStartTime
        self._timestampGetter = self._ptsTimestamp
        self._ptsPolicy = PTSPolicy.pts

    def _setUpServerTimestampGetter(self):
        """Set up server time as pts source"""
        self._timeShift = 0
        self._timestampGetter = time.time
        self._ptsPolicy = PTSPolicy.server

    def _checkPTSCorrectness(self) -> bool:
        """
        Heuristic function for package pts correctness check.

        If video is file then pts must be near 0.
        If video is stream then pts must be near current unix timestamp.
        """
        self._videoTimeBase = self._decoder.Timebase()
        pts = self._packetData.pts
        if pts is None:
            return False
        pts = pts * self._videoTimeBase
        if self._source == VideoSource.stream:
            if abs(pts - time.time()) > 60 * 60 * 24:
                return False
        elif self._source == VideoSource.file:
            if abs(pts) > 10**5:
                return False
        return True

    def _setUpFrameTimestampCalculator(self) -> PTSPolicy:
        """
        Choose frame timestamp calculator

        Return actual pts policy
        """
        self._avgFramerate = self._decoder.AvgFramerate()
        match self._ptsPolicy:
            case PTSPolicy.frameRate:
                logger.info("set up framerate pts")
                self._setUpFramerateTimestampGetter()
            case PTSPolicy.server:
                logger.info("set up server pts")
                self._setUpServerTimestampGetter()
            case PTSPolicy.pts:
                pts = self._packetData.pts
                if pts == nvc.NO_PTS:
                    raise PTSNotFound
                logger.info("set up video pts")
                self._setUpPTSTimestampGetter()
            case _:
                pts = self._packetData.pts
                if pts == nvc.NO_PTS or not self._checkPTSCorrectness():
                    if self._source == VideoSource.stream:
                        logger.info("set up server pts")
                        self._setUpServerTimestampGetter()
                    else:
                        logger.info("set up framerate pts")
                        self._setUpFramerateTimestampGetter()
                else:
                    logger.info("set up video pts")
                    self._setUpPTSTimestampGetter()
        return self._ptsPolicy

    def _ptsTimestamp(self) -> float:
        """Get frame timestamp by pts from package"""
        if (ts := self._packetData.pts) != nvc.NO_PTS:
            return ts * self._videoTimeBase + self._timeShift
        else:
            return self._lastTS + 1 / self._avgFramerate

    async def getFramerate(self) -> float:
        """Get video framerate"""
        if self._avgFramerate == 0.0:
            decoder = await self._getDecoder()
            self._avgFramerate = decoder.AvgFramerate()
        return self._avgFramerate

    @classmethod
    async def close(cls) -> None:
        """Shutdown executor and session"""
        if cls.executor:
            cls.executor.shutdown(wait=True)

    @property
    def processingParts(self) -> list[VideoSegment] | None:
        """
        Get success decoded parts

        Returns:
            all success processing parts
        """
        return self._processingParts

    @property
    def coreDuration(self) -> float | None:
        """Get core duration data. Return None if data is not available."""
        duration = self._decoder.Duration()
        if duration:
            return duration
        return None

    @property
    def duration(self) -> float | None:
        """Get video duration. Return None if data is not available"""
        duration = self._decoder.Duration()
        if duration:
            return duration
        numFrames = self.numFrames
        if numFrames is None:
            return None
        return numFrames / self._decoder.AvgFramerate()

    @property
    def numFrames(self) -> int | None:
        """
        Get video frame count.
        """
        if numFrames := self._decoder.NumFrames():
            pass
        elif self._state == DecoderState.completed:
            numFrames = self.decodedFrameCount
        else:
            return None
        return numFrames

    @property
    def bitrate(self) -> int:
        """
        Current stream bitrate
        """
        return self._decoder.Bitrate()

    @property
    def width(self) -> int:
        """
        Current frame width
        """
        return self._decoder.Width()

    @property
    def height(self) -> int:
        """
        Current frame height
        """
        return self._decoder.Height()

    @property
    def errorCount(self) -> int:
        """Get decoding error count"""
        return self._errorCount
