import asyncio
import tempfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from uuid import uuid4

import aiohttp
from luna3.common.base_api import makeRequest
from luna3.common.exceptions import LunaApiException


class VideoDownloadException(Exception):
    """Video opening exception"""


class LargeVideoException(VideoDownloadException):
    """Too large video"""

    def __init__(self, size: int):
        self.size = size


class VideoSavingTimeoutException(Exception):
    """Video saving timeout"""


class VideoSaver:
    """Context manager for video saving"""

    clientSession: aiohttp.ClientSession | None = None
    tmpStorage: tempfile.TemporaryDirectory[str]
    executor: ThreadPoolExecutor | None = None

    def __init__(self, url: str, maxVideoSize: int | None = None, totalTimeout: int | None = None):
        self.url = url
        self.loop = asyncio.get_running_loop()
        self.videoPath: Path | None = None
        self.maxVideoSize = maxVideoSize
        self._totalTimeout = totalTimeout

    @classmethod
    async def initialize(cls, timeouts: "ServiceTimeoutsSettings", tmpStorage: Path | str | None = None) -> None:  # type: ignore
        """Initialize video saver"""
        tmpStorage = tmpStorage or Path(__file__).parent.resolve()
        cls.clientSession = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=timeouts.totalTimeout,
                connect=timeouts.connectTimeout,
                sock_connect=timeouts.sockConnectTimeout,
                sock_read=timeouts.sockReadTimeout,
            ),
        )
        cls.tmpStorage = tempfile.TemporaryDirectory(dir=tmpStorage)
        cls.executor = ThreadPoolExecutor(max_workers=8)

    async def __aenter__(self) -> Path:
        """Context manager to handle video file saving and deletion"""
        kwargs = {
            "asyncRequest": True,
            "raiseError": True,
            "url": self.url,
            "method": "GET",
            "session": self.clientSession,
        }
        if self._totalTimeout is not None:
            kwargs["totalTimeout"] = self._totalTimeout
            kwargs["connectTimeout"] = kwargs["sockConnectTimeout"] = kwargs["sockReadTimeout"] = None
        try:
            response = await makeRequest(**kwargs)

            if self.maxVideoSize:
                # todo: use stream for video downloading and read header immediately
                try:
                    contentLength = int(response.headers.get("Content-Length", 0))
                except ValueError:
                    contentLength = 0
                if contentLength and contentLength > self.maxVideoSize:
                    raise LargeVideoException(contentLength)

            video = response.body
            if self.maxVideoSize and len(video) > self.maxVideoSize:
                raise LargeVideoException(len(video))
        except LunaApiException as err:
            exception = VideoSavingTimeoutException if err.statusCode == 504 else VideoDownloadException
            raise exception(f"status code: {err.statusCode}, {err.json["desc"]}") from err
        return await self.loop.run_in_executor(self.executor, self.save, video)

    async def __aexit__(self, *args: list[str]) -> None:
        await self.loop.run_in_executor(self.executor, self.remove)

    def save(self, video: bytes) -> Path:
        """
        Save video file

        Args:
            video (bytes): bytes to save
        """
        self.videoPath = Path(self.tmpStorage.name) / str(uuid4())
        self.videoPath.write_bytes(video)
        return self.videoPath

    def remove(self) -> None:
        """Remove saved video"""
        if self.videoPath:
            self.videoPath.unlink()

    @classmethod
    async def close(cls) -> None:
        """Shutdown executor and session"""
        if cls.executor:
            cls.executor.shutdown(wait=True)
        if cls.clientSession:
            await cls.clientSession.close()
